/* A solution to the HPC University Parallel Forest Fire challenge
 *
 * Simulates a forest fire in an 80x80 grid, with a 100% chance of a
 * tree catching fire if it is next to a burning tree, but this parameter
 * can be changed using the -p option, e.g:
 * mpirun -np 2 ./mpi -p .5
 * , which would use a 50% probability.
 *
 * Check the README for more information.
 *
 * Aaron Weeden, Shodor Education Foundation, November 2013
 */

/* import libraries */
#include <mpi.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>

/* begin program execution */
int main(int argc, char **argv) {

    /* declare variables */
    int rowCount = 80, columnCount = 80, maxBurnStepCount = 5;
    int myRank, processorCount;
    int myRowCount;
    int *trees, *newTrees, *topNeighborTrees, *bottomNeighborTrees,
        *allTrees, *treeCountsForEachProcess, *displs;
    int row, column, rank, treeIndex;
    bool stillBurning;
    double probabilityOfCatch = 1.0;
    char character;

    /* start up MPI */
    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &myRank);
    MPI_Comm_size(MPI_COMM_WORLD, &processorCount);

    /* get user input from the command line */
    while ((character = getopt(argc, argv, "p:")) != -1) {
        switch (character) {
            case 'p':
                probabilityOfCatch = atof(optarg);
                break;
            default:
                fprintf(stderr, "Usage: %s [-p probabilityOfCatch]\n",
                        argv[0]);
                exit(EXIT_FAILURE);
        }
    }

    /* assign myRowCount, round-robin style */
    myRowCount = rowCount / processorCount;
    if (myRank < rowCount % processorCount) {
        myRowCount++;
    }

    /* allocate memory for arrays */
    trees    = (int*)malloc(myRowCount*columnCount*sizeof(int));
    newTrees = (int*)malloc(myRowCount*columnCount*sizeof(int));
    topNeighborTrees    = (int*)malloc(columnCount*sizeof(int));
    bottomNeighborTrees = (int*)malloc(columnCount*sizeof(int));

    /* set up arrays for MPI_Allgatherv */
    if (myRank == 0) {
        allTrees = (int*)malloc(rowCount*columnCount*sizeof(int));
        treeCountsForEachProcess = (int*)malloc(processorCount*sizeof(int));
        displs = (int*)malloc(processorCount*sizeof(int));

        displs[0] = 0;
        for (rank = 0; rank < processorCount; rank++) {
            if (rank != 0) {
                displs[rank] = displs[rank-1]
                    + treeCountsForEachProcess[rank-1];
            }
            treeCountsForEachProcess[rank] = rowCount / processorCount;
            if (rank < rowCount % processorCount) {
                treeCountsForEachProcess[rank]++;
            }
            treeCountsForEachProcess[rank] *= columnCount;
        }
    }

    /* initialize arrays with 0's */
    for (row = 0; row < myRowCount; row++) {
        for (column = 0; column < columnCount; column++) {
            newTrees[row * columnCount + column] = 0;
        }
    }
    for (column = 0; column < columnCount; column++) {
        topNeighborTrees[column] = bottomNeighborTrees[column] = 0;
    }

    /* light a random tree of Rank 0 */
    srandom(time(NULL));
    if (myRank == 0) {
        newTrees[myRowCount*columnCount*random()/RAND_MAX] = 1;
    }

    /* loop until no trees are left burning */
    stillBurning = true;
    while (stillBurning) {

        /* advance the grid from the last time step */
        for (row = 0; row < myRowCount; row++) {
            for (column = 0; column < columnCount; column++) {
                treeIndex = row * columnCount + column;
                trees[treeIndex] = newTrees[treeIndex];
            }
        }

        /* Rank 0 prints the grid */
        MPI_Gatherv(trees, myRowCount*columnCount, MPI_INT, allTrees, 
                treeCountsForEachProcess, displs, MPI_INT, 0, MPI_COMM_WORLD);
        if (myRank == 0) {
            for (row = 0; row < rowCount; row++) {
                for (column = 0; column < columnCount; column++) {
                    printf("%d", allTrees[row * columnCount + column]);
                }
                printf("\n");
            }
            printf("\n");
        }

        /* send/receive neighbor rows on the boundaries */
        if (myRank > 0) {
            MPI_Send(&(trees[0]), columnCount, MPI_INT, myRank-1, 0,
                    MPI_COMM_WORLD);
        }
        if (myRank < processorCount-1) {
            MPI_Send(&(trees[(myRowCount-1)*columnCount]), columnCount, 
                    MPI_INT, myRank+1, 0, MPI_COMM_WORLD);
        }
        if (myRank > 0) {
            MPI_Recv(topNeighborTrees, columnCount, MPI_INT, myRank-1, 0,
                    MPI_COMM_WORLD, MPI_STATUS_IGNORE);
        }
        if (myRank < processorCount-1) {
            MPI_Recv(bottomNeighborTrees, columnCount, MPI_INT, myRank+1, 0,
                    MPI_COMM_WORLD, MPI_STATUS_IGNORE);
        }

        /* spread the fire */
        for (row = 0; row < myRowCount; row++) {
            for (column = 0; column < columnCount; column++) {
                treeIndex = row * columnCount + column;
                if (trees[treeIndex] == 0) {
                    if ((column > 0
                                && trees[treeIndex-1] > 0
                                && trees[treeIndex-1] < maxBurnStepCount)
                            || (row == 0
                                && topNeighborTrees[column] > 0
                                && topNeighborTrees[column]
                                < maxBurnStepCount) 
                            || (row > 0
                                && trees[treeIndex-columnCount] > 0
                                && trees[treeIndex-columnCount]
                                < maxBurnStepCount)
                            || (column < columnCount-1
                                && trees[treeIndex+1] > 0
                                && trees[treeIndex+1] < maxBurnStepCount)
                            || (row == myRowCount-1
                                && bottomNeighborTrees[column] > 0
                                && bottomNeighborTrees[column]
                                > maxBurnStepCount)
                            || (row < myRowCount-1
                                && trees[treeIndex+columnCount] > 0
                                && trees[treeIndex+columnCount]
                                < maxBurnStepCount)) {
                                    if ((double)random()/RAND_MAX
                                            < probabilityOfCatch) {
                                        newTrees[treeIndex] = 1;
                                    }
                                }
                } else if (trees[treeIndex] < maxBurnStepCount) {
                    newTrees[treeIndex] = trees[treeIndex] + 1;
                } else {
                    newTrees[treeIndex] = trees[treeIndex];
                }
            }
        }

        /* check for whether fire has stopped */
        stillBurning = false;
        for (row = 0; !stillBurning && row < myRowCount; row++) {
            for (column = 0; column < columnCount; column++) {

                treeIndex = row * columnCount + column;

                if (newTrees[treeIndex] > 0
                        && newTrees[treeIndex] < maxBurnStepCount) {
                    stillBurning = true;
                }
            }
        }
        MPI_Allreduce(&stillBurning, &stillBurning, 1, MPI_INT, MPI_LOR,
                MPI_COMM_WORLD);
    }

    /* free memory for arrays */
    if (myRank == 0) {
        free(displs);
        free(treeCountsForEachProcess);
        free(allTrees);
    }
    free(bottomNeighborTrees);
    free(topNeighborTrees);
    free(newTrees);
    free(trees);

    /* finish up */
    MPI_Finalize();
    return 0;
}
